using System;
using System.Collections.Generic;
using System.Text;

namespace IntSight.Controls
{
    public partial class CodeEditor
    {
        public struct Position : IComparable<Position>, IEquatable<Position>
        {
            public static readonly Position Zero =
                new Position();
            public static readonly Position Infinite =
                new Position(int.MaxValue, int.MaxValue);

            internal int line, column;

            public Position(int line, int column)
            {
                this.line = line;
                if (column < 0)
                    this.column = 0;
                else
                    this.column = column;
            }

            public int Line
            {
                get { return line; }
                set { line = value; }
            }

            public int Column
            {
                get { return column; }
                set { column = value; }
            }

            internal Position ChangeLine(int value)
            {
                return new Position(value, column);
            }

            public override bool Equals(object obj)
            {
                if (obj != null && obj.GetType() == typeof(Position))
                {
                    Position pos = (Position)obj;
                    return line == pos.line && column == pos.column;
                }
                else
                    return false;
            }

            public override int GetHashCode()
            {
                return line.GetHashCode() ^ column.GetHashCode();
            }

            public override string ToString()
            {
                return String.Format("{0}, {1}", line, column);
            }

            public Position RangeDecrement(Position p1, Position p2)
            {
                if (p1.IsGreaterThan(p2))
                {
                    Position temp = p1; p1 = p2; p2 = temp;
                }
                if (this.IsGreaterEqual(p2))
                {
                    if (line == p2.line)
                        column = column - p2.column + p1.column;
                    line -= p2.line - p1.line;
                }
                return this;
            }

            /// <summary>Efficient test for position inequality.</summary>
            /// <param name="other">Position to compare with.</param>
            /// <returns>True if positions are not equals.</returns>
            /// <remarks>Operators pass both parameters by value.</remarks>
            internal bool NotEquals(Position other)
            {
                return this.line != other.line || this.column != other.column;
            }

            internal bool IsLesserThan(Position p)
            {
                return line < p.line || line == p.line && column < p.column;
            }

            internal bool IsGreaterThan(Position p)
            {
                return line > p.line || line == p.line && column > p.column;
            }

            internal bool IsLesserEqual(Position p)
            {
                return line < p.line || line == p.line && column <= p.column;
            }

            internal bool IsGreaterEqual(Position p)
            {
                return line > p.line || line == p.line && column >= p.column;
            }

            public static bool operator ==(Position p1, Position p2)
            {
                return p1.line == p2.line && p1.column == p2.column;
            }

            public static bool operator !=(Position p1, Position p2)
            {
                return p1.line != p2.line || p1.column != p2.column;
            }

            public static bool operator <(Position p1, Position p2)
            {
                return p1.line < p2.line || p1.line == p2.line && p1.column < p2.column;
            }

            public static bool operator >(Position p1, Position p2)
            {
                return p1.line > p2.line || p1.line == p2.line && p1.column > p2.column;
            }

            public static bool operator <=(Position p1, Position p2)
            {
                return p1.line < p2.line || p1.line == p2.line && p1.column <= p2.column;
            }

            public static bool operator >=(Position p1, Position p2)
            {
                return p1.line > p2.line || p1.line == p2.line && p1.column >= p2.column;
            }

            #region IComparable<Position> members.

            public int CompareTo(Position other)
            {
                int result = line.CompareTo(other.line);
                return result != 0 ? result : column.CompareTo(other.column);
            }

            #endregion

            #region IEquatable<Position> members.

            public bool Equals(Position other)
            {
                return this.line == other.line && this.column == other.column;
            }

            #endregion
        }
    }

    public delegate void SelectionChangedEventHandler(
        object sender, SelectionChangedEventArgs e);

    public class SelectionChangedEventArgs : EventArgs
    {
        private CodeEditor.Position originalFrom, originalTo;

        public SelectionChangedEventArgs(
            CodeEditor.Position originalFrom, CodeEditor.Position originalTo)
        {
            this.originalFrom = originalFrom;
            this.originalTo = originalTo;
        }

        internal void Reset(CodeEditor.Position originalFrom, CodeEditor.Position originalTo)
        {
            this.originalFrom = originalFrom;
            this.originalTo = originalTo;
        }

        public CodeEditor.Position OriginalFrom { get { return originalFrom; } }
        public CodeEditor.Position OriginalTo { get { return originalTo; } }
    }
}
