using System;
using System.Collections.Generic;

namespace IntSight.Controls
{
    using IntSight.Controls.CodeModel;
    using Rsc = IntSight.Controls.Properties.Resources;

    public partial class CodeEditor
    {
        /// <summary>The common abstract base for all undo commands.</summary>
        private abstract class UndoAction
        {
            protected Position start, end;

            public UndoAction(Position start, Position end)
            {
                this.start = start;
                this.end = end;
            }

            /// <summary>An attempt to merge or group two editor commands.</summary>
            /// <param name="another">The newest command.</param>
            /// <returns>Returns true when successful.</returns>
            /// <remarks>
            /// When <c>TryMerge</c> succeeds, <c>this</c> is updated to contain
            /// all relevant information from <c>another</c>.
            /// </remarks>
            public virtual bool TryMerge(UndoAction another)
            {
                return false;
            }

            public virtual int FirstLine
            {
                get { return Math.Min(start.line, end.line); }
            }

            public virtual bool TryClearIndent(ICodeModel model)
            {
                return false;
            }

            /// <summary>Undo the represented editor command.</summary>
            /// <param name="model">The document being edited.</param>
            public abstract void Undo(ICodeModel model);

            /// <summary>Repeats the action of the represented editor command.</summary>
            /// <param name="model">The document being edited.</param>
            public abstract void Redo(ICodeModel model);

            public abstract string Description { get; }
            public abstract int ByteCost { get; }
        }

        private sealed class DeleteAction : UndoAction
        {
            private string text;

            public DeleteAction(Position start, Position end, string text)
                : base(start, end)
            {
                this.text = text;
            }

            public bool MatchesInsertion(Position iEnd)
            {
                return iEnd.Equals(this.end) &&
                    start.line == end.line && start.column == end.column - 1;
            }

            public override bool TryMerge(UndoAction another)
            {
                DeleteAction other = another as DeleteAction;
                if (other == null)
                    return false;
                else if (this.start.Equals(other.end))
                {
                    this.text = other.text + this.text;
                    this.start = other.start;
                    return true;
                }
                else if (this.start.Equals(other.start) && other.text.Length == 1)
                {
                    this.text += other.text;
                    this.end.column++;
                    return true;
                }
                else
                    return false;
            }

            public override void Redo(ICodeModel model)
            {
                model.DeleteRange(start, end);
            }

            public override void Undo(ICodeModel model)
            {
                model.MoveTo(start, false);
                model.InsertRange(start, text);
            }

            public override string Description { get { return Rsc.UndoCmdDeletion; } }
            public override int ByteCost { get { return 20 + text.Length; } }
        }

        private sealed class DragMoveAction : UndoAction
        {
            private Position startText, endText;
            private string text;

            public DragMoveAction(Position start, Position end,
                Position startText, Position endText, string text)
                : base(start, end)
            {
                this.startText = startText;
                this.endText = endText;
                this.text = text;
            }

            public override int FirstLine
            {
                get { return Math.Min(Math.Min(base.FirstLine, startText.line), endText.line); }
            }

            public override void Undo(ICodeModel model)
            {
                if (startText.IsGreaterEqual(end))
                {
                    model.InsertRange(start, text);
                    model.DeleteRange(startText, endText);
                }
                else
                {
                    model.DeleteRange(startText, endText);
                    model.InsertRange(start, text);
                }
                model.MoveTo(start, false);
                model.MoveTo(end, true);
            }

            public override void Redo(ICodeModel model)
            {
                if (startText.IsGreaterEqual(end))
                {
                    model.InsertRange(startText, text);
                    model.DeleteRange(start, end);
                }
                else
                {
                    model.DeleteRange(start, end);
                    model.InsertRange(startText, text);
                }
            }

            public override string Description { get { return Rsc.UndoCmdDragMove; } }
            public override int ByteCost { get { return 36 + text.Length; } }
        }

        private abstract class BaseCommentAction : UndoAction
        {
            protected List<Position> insertPoints;
            protected string lineComment;

            public BaseCommentAction(string lineComment)
                : base(Position.Infinite, Position.Infinite)
            {
                this.insertPoints = new List<Position>();
                this.lineComment = lineComment;
            }

            public void Add(Position position)
            {
                insertPoints.Add(position);
                if (position.IsLesserThan(start))
                    start = position;
            }

            public sealed override int ByteCost { get { return 20 + 8 * insertPoints.Count; } }
        }

        private sealed class UncommentAction : BaseCommentAction
        {
            public UncommentAction(string lineComment) : base(lineComment) { }

            public override void Undo(ICodeModel model)
            {
                foreach (Position p in insertPoints)
                    model.InsertRange(p, lineComment);
            }

            public override void Redo(ICodeModel model)
            {
                foreach (Position p in insertPoints)
                    model.DeleteRange(p, new Position(p.line, p.column + lineComment.Length));
            }

            public override string Description { get { return Rsc.UndoCmdUncomment; } }
        }

        private sealed class CommentAction : BaseCommentAction
        {
            public CommentAction(string lineComment) : base(lineComment) { }

            public override void Undo(ICodeModel model)
            {
                for (int i = insertPoints.Count - 1; i >= 0; i--)
                {
                    Position p = insertPoints[i];
                    model.DeleteRange(p, new Position(p.line, p.column + lineComment.Length));
                }
            }

            public override void Redo(ICodeModel model)
            {
                foreach (Position p in insertPoints)
                    model.InsertRange(p, lineComment);
            }

            public override string Description { get { return Rsc.UndoCmdComment; } }
        }

        private abstract class InsertAction : UndoAction
        {
            private string text = String.Empty;

            protected InsertAction(Position start, Position end)
                : base(start, end) { }

            public override void Redo(ICodeModel model)
            {
                model.InsertRange(start, text);
                text = String.Empty;
            }

            public override void Undo(ICodeModel model)
            {
                text = model.TextRange(start, end);
                model.DeleteRange(start, end);
            }

            public sealed override int ByteCost { get { return 20 + text.Length; } }
        }

        private sealed class TypeAction : InsertAction
        {
            public TypeAction(Position start, Position end)
                : base(start, end) { }

            public override bool TryMerge(UndoAction another)
            {
                TypeAction other = another as TypeAction;
                if (other == null)
                {
                    // Check if new action is a backspace.
                    DeleteAction delete = another as DeleteAction;
                    if (delete != null && delete.MatchesInsertion(this.end))
                    {
                        end.column--;
                        return true;
                    }
                    else
                        return false;
                }
                else if (this.end.Equals(other.start))
                {
                    this.end = other.end;
                    return true;
                }
                else
                    return false;
            }

            public override bool TryClearIndent(ICodeModel model)
            {
                if (model.Current.Equals(this.end))
                {
                    Position newEnd = new Position(this.end.line, 0);
                    if (this.start.IsLesserThan(newEnd))
                    {
                        this.end = newEnd;
                        return true;
                    }
                }
                return false;
            }

            public override string Description { get { return Rsc.UndoCmdTyping; } }
        }

        private sealed class DragCopyAction : InsertAction
        {
            public DragCopyAction(Position start, Position end)
                : base(start, end) { }

            public override string Description { get { return Rsc.UndoCmdDragCopy; } }
        }

        private sealed class PasteAction : InsertAction
        {
            public PasteAction(Position start, Position end)
                : base(start, end) { }

            public override string Description { get { return Rsc.UndoCmdPaste; } }
        }

        private sealed class CaseChangeAction : UndoAction
        {
            private string text;
            private bool toUpper;

            public CaseChangeAction(Position start, Position end, string text, bool toUpper)
                : base(start, end)
            {
                this.text = text;
                this.toUpper = toUpper;
            }

            public override void Undo(ICodeModel model)
            {
                model.DeleteRange(start, end);
                model.InsertRange(start, text);
                text = String.Empty;
            }

            public override void Redo(ICodeModel model)
            {
                text = model.TextRange(start, end);
                model.DeleteRange(start, end);
                model.InsertRange(start,
                    toUpper ? text.ToUpperInvariant() : text.ToLowerInvariant());
            }

            public override string Description { get { return Rsc.UndoCmdTransform; } }
            public override int ByteCost { get { return 21 + text.Length; } }
        }

        private class ReplaceAction : UndoAction
        {
            private string text;

            public ReplaceAction(Position start, Position end, string text)
                : base(start, end)
            {
                this.text = text;
            }

            public override void Undo(ICodeModel model)
            {
                string s = model.TextRange(start, end);
                model.DeleteRange(start, end);
                model.InsertRange(start, text);
                end = model.Current;
                text = s;
            }

            public override void Redo(ICodeModel model)
            {
                Undo(model);
            }

            public override string Description { get { return Rsc.UndoCmdReplace; } }
            public override int ByteCost { get { return 20 + text.Length; } }
        }

        private sealed class SnippetAction : ReplaceAction
        {
            public SnippetAction(Position start, Position end, string text)
                : base(start, end, text) { }

            public override string Description { get { return Rsc.UndoCmdSnippet; } }
        }

        private sealed class IndentAction : UndoAction
        {
            private int amount;

            public IndentAction(Position start, Position end, int amount)
                : base(start, end) { this.amount = amount; }

            public override void Redo(ICodeModel model)
            {
                model.IndentRange(start, end, amount, null);
            }

            public override void Undo(ICodeModel model)
            {
                model.UnindentRange(start, end, amount);
            }

            public override string Description { get { return Rsc.UndoCmdIndent; } }
            public override int ByteCost { get { return 20; } }
        }

        private sealed class UnindentAction : UndoAction
        {
            private static int[] emptyList = new int[0];
            private int amount;
            private int[] exceptions;

            public UnindentAction(Position start, Position end, int amount,
                params int[] exceptions)
                : base(start, end)
            {
                this.amount = amount;
                this.exceptions = exceptions ?? emptyList;
            }

            public override void Redo(ICodeModel model)
            {
                model.UnindentRange(start, end, amount);
            }

            public override void Undo(ICodeModel model)
            {
                model.IndentRange(start, end, amount, exceptions);
            }

            public override string Description { get { return Rsc.UndoCmdOutdent; } }
            public override int ByteCost { get { return 24 + exceptions.Length; } }
        }
    }
}